from __future__ import annotations

import typing
from pydoc import locate
from typing import Any

from injector import inject, singleton

from open_ticket_ai.core import AppConfig
from open_ticket_ai.core.renderable.renderable import Renderable, RenderableConfig
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig
from open_ticket_ai.core.pipeline.pipe_context import PipeContext
from open_ticket_ai.core.template_rendering.template_renderer import TemplateRenderer


def _locate(use: str) -> type:
    if ":" in use:
        m, c = use.split(":", 1)
        use = f"{m}.{c}"
    use_class = locate(use)
    if use_class is None:
        raise ValueError(f"Cannot locate class '{use}'")
    return typing.cast(type, locate(use))


@singleton
class RenderableFactory:
    @inject
    def __init__(
        self,
        template_renderer: TemplateRenderer,
        app_config: AppConfig,
        registerable_configs: list[RenderableConfig],
        logger_factory: LoggerFactory,
    ):
        self._logger = logger_factory.get_logger(self.__class__.__name__)
        self._template_renderer = template_renderer
        self._registerable_configs = registerable_configs
        self._app_config = app_config
        self._logger_factory = logger_factory

    def create_pipe(self, config_raw: PipeConfig, scope: PipeContext) -> Pipe:
        self._logger.debug(f"Creating pipe with config id: {config_raw.id}")
        self._logger.info(f"Creating pipe '{config_raw.id}'")
        rendered_instance = self.__render(config_raw, scope)
        if not isinstance(rendered_instance, Pipe):
            raise TypeError(f"Registerable with id '{config_raw.id}' is not a Pipe")
        return rendered_instance

    def create_trigger(self, trigger_config_raw: RenderableConfig, scope: PipeContext) -> Renderable:
        self._logger.debug(f"Creating trigger with config id: {trigger_config_raw.id}")
        self._logger.info(f"Creating trigger '{trigger_config_raw.id}'")
        return self.__render(trigger_config_raw, scope)

    def __create_service_instance(self, registerable_config_raw: RenderableConfig, scope: PipeContext) -> Renderable:
        return self.__render(registerable_config_raw, scope)

    def __render(self, registerable_config_raw: RenderableConfig, scope: PipeContext):
        rendered_params = self._template_renderer.render_recursive(registerable_config_raw.params, scope)
        registerable_config_raw.params = rendered_params
        return self.__create_renderable_instance(registerable_config_raw, scope)

    def __create_renderable_instance(self, rendered_config: RenderableConfig, scope: PipeContext) -> Renderable:
        cls: type = _locate(rendered_config.use)
        if not issubclass(cls, Renderable):
            raise TypeError(f"Class '{rendered_config.use}' is not a {Renderable.__class__.__name__}")

        kwargs: dict[str, Any] = {
            "factory": self,
            "app_config": self._app_config,
            "logger_factory": self._logger_factory,
            "config": rendered_config,
        }
        kwargs.update(self.__resolve_injects(rendered_config.injects, scope))
        return cls(**kwargs)

    def __resolve_injects(self, injects: dict[str, str], scope: PipeContext) -> dict[str, Renderable]:
        return {param: self.__resolve_by_id(ref, scope) for param, ref in injects.items()}

    def __resolve_by_id(self, service_id: str, scope: PipeContext) -> Any:
        matching_config = next(
            (config for config in self._registerable_configs
             if RenderableConfig.model_validate(config).id == service_id),
            None
        )
        if matching_config is None:
            raise KeyError(service_id)
        return self.__create_service_instance(matching_config, scope)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict

from open_ticket_ai.core.renderable.renderable import Renderable
from ..logging_iface import LoggerFactory
from .pipe_config import PipeConfig, PipeResult
from .pipe_context import PipeContext


class ParamsModel(BaseModel):
    model_config = ConfigDict(extra="allow")


class Pipe(Renderable, ABC):
    def __init__(self, config: PipeConfig, logger_factory: LoggerFactory, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._config = PipeConfig.model_validate(config.model_dump())
        self._logger = logger_factory.get_logger(self.__class__.__name__)

    def _have_dependent_pipes_been_run(self, context: PipeContext) -> bool:
        return all(context.has_succeeded(dependency_id) for dependency_id in self._config.depends_on)

    async def _process_and_save(self, context: PipeContext) -> PipeContext:
        pipe_result = await self._process()
        return context.with_pipe_result(self._config.id, pipe_result)

    async def process(self, context: PipeContext) -> PipeContext:
        self._logger.info(f"Processing pipe '{self._config.id}'")
        if self._config.should_run and self._have_dependent_pipes_been_run(context):
            self._logger.info(f"Pipe '{self._config.id}' is running.")

            return await self._process_and_save(context)
        self._logger.info(f"Skipping pipe '{self._config.id}'.")
        return context


    @abstractmethod
    async def _process(self) -> PipeResult:
        pass



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.renderable.renderable import EmptyModel
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult
from open_ticket_ai.core.ticket_system_integration.ticket_system_service import TicketSystemService
from open_ticket_ai.core.ticket_system_integration.unified_models import UnifiedNote


class AddNoteParams(BaseModel):
    ticket_id: str | int
    note: UnifiedNote



class AddNotePipeConfig(PipeConfig):
    params = AddNoteParams


class AddNotePipe(Pipe):
    def __init__(
        self,
        ticket_system: TicketSystemService,
        config: AddNotePipeConfig,
        logger_factory: LoggerFactory,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(config, logger_factory=logger_factory)
        self._config = AddNotePipeConfig.model_validate(config.model_dump())
        self._ticket_system = ticket_system

    async def _process(self) -> PipeResult:
        ticket_id_str = str(self._config.params.ticket_id)
        await self._ticket_system.add_note(ticket_id_str, self._config.params.note)
        return PipeResult(success=True, failed=False, data=EmptyModel())



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult
from open_ticket_ai.core.ticket_system_integration.ticket_system_service import TicketSystemService
from open_ticket_ai.core.ticket_system_integration.unified_models import UnifiedTicket


class UpdateTicketParams(BaseModel):
    ticket_id: str
    updated_ticket: UnifiedTicket


class UpdateTicketPipeConfig(PipeConfig):
    params: UpdateTicketParams


class UpdateTicketPipe(Pipe):
    def __init__(
        self,
        ticket_system: TicketSystemService,
        config: UpdateTicketPipeConfig,
        logger_factory: LoggerFactory | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if logger_factory is None:
            raise ValueError("logger_factory is required")
        super().__init__(config, logger_factory=logger_factory)
        self._config = UpdateTicketPipeConfig.model_validate(config.model_dump())
        self._ticket_system = ticket_system

    async def _process(self) -> PipeResult:
        success = await self._ticket_system.update_ticket(
            ticket_id=self._config.params.ticket_id,
            updates=self._config.params.updated_ticket,
        )
        if not success:
            return PipeResult(
                success=False,
                message="Failed to update ticket",
            )
        return PipeResult(success=True)



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field


class TemplateRendererConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    type: str = Field(..., description="Type of template renderer")


class JinjaRendererConfig(TemplateRendererConfig):
    type: Literal["jinja"] = "jinja"


class MustacheRendererConfig(TemplateRendererConfig):
    type: Literal["mustache"] = "mustache"


SpecificTemplateRendererConfig = Annotated[JinjaRendererConfig | MustacheRendererConfig, Field(discriminator="type")]



import asyncio
from abc import ABC, abstractmethod
from typing import Any, Protocol

from open_ticket_ai.core.renderable.renderable import Renderable
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.orchestration.orchestrator_config import TriggerConfig


class PipeRunnerObserver(Protocol):
    async def on_trigger_fired(self) -> None: ...


class Trigger(Renderable, ABC):
    def __init__(self, config: TriggerConfig, logger_factory: LoggerFactory, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.trigger_def = config
        self._observers: list[PipeRunnerObserver] = []
        self._running = False
        self._logger = logger_factory.get_logger(self.__class__.__name__)

    def attach(self, observer: PipeRunnerObserver) -> None:
        if observer not in self._observers:
            self._observers.append(observer)

    def detach(self, observer: PipeRunnerObserver) -> None:
        if observer in self._observers:
            self._observers.remove(observer)

    def notify(self) -> None:
        for obs in self._observers:
            asyncio.create_task(obs.on_trigger_fired())

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field

from open_ticket_ai.core.config.logging_config import LoggingDictConfig
from open_ticket_ai.core.renderable.renderable import RenderableConfig
from open_ticket_ai.core.orchestration.orchestrator_config import OrchestratorConfig


class InfrastructureConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    logging: LoggingDictConfig = Field(default_factory=LoggingDictConfig)
    default_template_renderer: str


class RawOpenTicketAIConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    plugins: list[str] = Field(default_factory=lambda: [])
    infrastructure: InfrastructureConfig = Field(default_factory=InfrastructureConfig)
    services: list[RenderableConfig] = Field(default_factory=lambda: [])
    orchestrator: OrchestratorConfig = Field(default_factory=OrchestratorConfig)



from typing import Any

from open_ticket_ai.core.renderable.renderable_factory import RenderableFactory
from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult
from open_ticket_ai.core.pipeline.pipe_context import PipeContext


class CompositePipeConfig(PipeConfig):
    steps: list[PipeConfig]


class CompositePipe(Pipe):
    def __init__(
            self,
            config: CompositePipeConfig,
            factory: RenderableFactory | None = None,
            logger_factory: LoggerFactory | None = None,
            *args: Any,
            **kwargs: Any,
    ) -> None:
        super().__init__(config, logger_factory=logger_factory)
        self._logger = logger_factory.get_logger(self.__class__.__name__)
        self.config = CompositePipeConfig.model_validate(config.model_dump())
        self._factory = factory

    def _build_pipe_from_step_config(self, step_config: PipeConfig, context: PipeContext) -> Pipe:
        if self._factory is None:
            raise ValueError("RenderableFactory is required but not provided to CompositePipe")
        return self._factory.create_pipe(step_config, context)

    async def _process_steps(self, context: PipeContext) -> list[PipeResult]:
        results: list[PipeResult] = []
        for step_config_raw in self.config.steps or []:
            context.model_copy(update={"parent": context})
            step_pipe = self._build_pipe_from_step_config(step_config_raw, context)
            context = await step_pipe.process(context)
            if step_config_raw.id in context.pipe_results:
                results.append(context.pipe_results[step_config_raw.id])
        return results

    async def _process_and_save(self, context: PipeContext) -> PipeContext:
        pipe_results = await self._process_steps(context)
        pipe_result = PipeResult.union(*pipe_results)
        return context.with_pipe_result(self.config.id, pipe_result)

    async def _process(self) -> PipeResult:
        raise NotImplementedError()



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field

from open_ticket_ai.core.renderable.renderable import RenderableConfig, EmptyModel
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig


class TriggerConfig(RenderableConfig):
    pass


class RunnerDefinition(BaseModel):
    id: str | None = None
    on: list[TriggerConfig]
    run: PipeConfig
    params: EmptyModel = Field(default_factory=EmptyModel)

    model_config = ConfigDict(populate_by_name=True)

    @property
    def pipe_id(self) -> str:
        if self.id is not None:
            return self.id
        if self.run.id is not None:
            return self.run.id
        return ""


class OrchestratorConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    runners: list[RunnerDefinition] = Field(default_factory=list)



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from open_ticket_ai.core.pipeline.pipe_config import PipeResult


class PipeContext(BaseModel):
    model_config = ConfigDict(frozen=True)
    pipe_results: dict[str, PipeResult] = Field(default_factory=dict)
    params: dict[str, Any] = Field(default_factory=dict)
    parent: PipeContext | None = Field(default=None, exclude=True)

    def has_succeeded(self, pipe_id: str) -> bool:
        pipe_result = self.pipe_results.get(pipe_id)
        return pipe_result is not None and pipe_result.success

    def with_pipe_result(self, pipe_id: str, pipe_result: PipeResult) -> "PipeContext":
        new_pipes = {**self.pipes, pipe_id: pipe_result}
        return self.model_copy(update={"pipes": new_pipes})



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult
from open_ticket_ai.core.ticket_system_integration.ticket_system_service import TicketSystemService
from open_ticket_ai.core.ticket_system_integration.unified_models import TicketSearchCriteria, UnifiedTicket


class FetchTicketsParams(BaseModel):
    ticket_search_criteria: TicketSearchCriteria | None = None


class FetchTicketsPipeResultData(BaseModel):
    fetched_tickets: list[UnifiedTicket]


class FetchTicketsPipeConfig(PipeConfig):
    params: FetchTicketsParams


class FetchTicketsPipe(Pipe):
    def __init__(
        self,
        ticket_system: TicketSystemService,
        config: FetchTicketsPipeConfig,
        logger_factory: LoggerFactory | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if logger_factory is None:
            raise ValueError("logger_factory is required")
        super().__init__(config, logger_factory=logger_factory)
        self._config = FetchTicketsPipeConfig.model_validate(config.model_dump())
        self._ticket_system = ticket_system

    async def _process(self) -> PipeResult:
        search_criteria = self._config.params.ticket_search_criteria
        tickets = await self._ticket_system.find_tickets(search_criteria)
        return PipeResult(
            success=True,
            failed=False,
            data=FetchTicketsPipeResultData(fetched_tickets=tickets),
        )



from typing import Any

from pydantic import BaseModel

from open_ticket_ai.core.logging_iface import LoggerFactory
from open_ticket_ai.core.pipeline.pipe import Pipe
from open_ticket_ai.core.pipeline.pipe_config import PipeConfig, PipeResult


class ExpressionParams(BaseModel):
    expression: str


class ExpressionPipeResultData(BaseModel):
    value: Any


class ExpressionPipeConfig(PipeConfig):
    params: ExpressionParams


class ExpressionPipe(Pipe):
    def __init__(
            self,
            config: ExpressionPipeConfig,
            logger_factory: LoggerFactory | None = None,
            *args: Any,
            **kwargs: Any,
    ) -> None:
        if logger_factory is None:
            raise ValueError("logger_factory is required")
        super().__init__(config, logger_factory=logger_factory)
        self.config = ExpressionPipeConfig.model_validate(config.model_dump())

    async def _process(self) -> PipeResult:
        return PipeResult(success=True, failed=False,
                          data=ExpressionPipeResultData(value=self.config.params.expression))



import os
from pathlib import Path

import injector
import yaml
from injector import singleton

from open_ticket_ai.core import AppConfig
from open_ticket_ai.core.config.config_models import RawOpenTicketAIConfig
from open_ticket_ai.core.logging_iface import LoggerFactory


@singleton
class ConfigLoader:
    @injector.inject
    def __init__(self, app_config: AppConfig, logger_factory: LoggerFactory):
        self.app_config = app_config
        self._logger = logger_factory.get_logger(self.__class__.__name__)

    def load_config(self, config_path: os.PathLike | None = None) -> RawOpenTicketAIConfig:
        if config_path is None:
            env_path = os.getenv(self.app_config.config_env_var)
            config_path = env_path if env_path else self.app_config.get_default_config_path()

        if not Path(config_path).exists():
            raise FileNotFoundError(
                f"Config file not found at {config_path}. "
                f"Create a config file at this path, provide a valid path, "
                f"or set the {self.app_config.config_env_var} environment variable."
            )

        with open(config_path) as file:
            yaml_content = yaml.safe_load(file)
            if yaml_content is None or self.app_config.config_yaml_root_key not in yaml_content:
                raise ValueError(f"Config file must contain '{self.app_config.config_yaml_root_key}' root key")
            config_dict = yaml_content[self.app_config.config_yaml_root_key]
            raw_otai_config = RawOpenTicketAIConfig.model_validate(config_dict)
        self._logger.info(f"Loaded config from {config_path}")
        return raw_otai_config


def load_config(
    config_path: str | os.PathLike[str] | None = None, app_config: AppConfig | None = None
) -> RawOpenTicketAIConfig:
    if app_config is None:
        app_config = AppConfig()

    loader = ConfigLoader(app_config)
    return loader.load_config(config_path)
